/*
 * Copyright (c) 2013 by Piero Orsoni (www.opbyte.it)
 * Based on the "sash" project by David I. Bell
 *
 * This file is part of cbmsh, a unix-like shell for commodore computers.
 *
 * Cbmsh is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cbmsh is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with cbmsh.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "cbmsh.h"
#include <conio.h>
#include <errno.h>
#include <peekpoke.h>

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

#define GREETING "cbmsh " STR(CBMSH_VERSION) "\n"
#define RC_FILENAME "cbmshrc"

// The special max argument value which means that there is no limit to the number of arguments on the command line.
#define	INFINITE_ARGS	255


// One entry of the command table.
typedef struct {
	const char *name;
	void (*func)(int argc, const char **argv);
	char minArgs;
	char maxArgs;
//	const char *usage;
} CommandEntry;


// The table of built-in commands. This is terminated wih an entry containing NULL values.
static const CommandEntry	commandEntryTable[] = {
	#ifndef CBMSH_MINIMAL
	{
		"alias",	do_alias,	1,	INFINITE_ARGS,
//		"[name [command]]"
	},
	#endif
	{
		"cd",		do_cd,		1,	2,
//		"[dirName]"
	},
	#ifndef CBMSH_MINIMAL
	{
		"clear",		do_clear,		1,	1,
//		""
	},
	#endif
	#ifndef CBMSH_MINIMAL
	{
		"cmp",		do_cmp,		3,	3,
//		"fileName1 fileName2"
	},
	#endif
	{
		"cp",		do_cp,		3,	INFINITE_ARGS,
//		"srcName ... destName"
	},
/*
	{
		"dd",		do_dd,		3,	INFINITE_ARGS,
//		"if=name of=name [bs=n] [count=n] [skip=n] [seek=n]"
	},
*/
	{
		"dos",		do_dos,	2,	INFINITE_ARGS,
//		"command ..."
	},
	{
		"echo",	do_echo,	1,	INFINITE_ARGS,
//		"[args] ..."
	},
/*
	{
		"ed",		do_ed,		1,	2,
//		"[fileName]"
	},
*/
	{
		"exit",		do_exit,	1,	1,
//		""
	},
/*
	{
		"file",	do_file,	1,	INFINITE_ARGS,
//		"fileName ..."
	},
	{
		"find",	do_find,	2,	INFINITE_ARGS,
//		"dirName [-xdev] [-type chars] [-name pattern] [-size minSize]"
	},
*/
	#ifndef CBMSH_MINIMAL
	{
		"free",		do_free,		1,	1,
//		""
	},
	#endif
/*
	{
		"grep",	do_grep,	3,	INFINITE_ARGS,
//		"[-in] word fileName ..."
	},
*/
	{
		"help",		do_help,	1,	1,
//		"[word]"
	},

	{
		"ls",		do_ls,		1,	INFINITE_ARGS,
//		"[-lu] fileName ..."
	},

	{
		"mkdir",	do_mkdir,	2,	INFINITE_ARGS,
//		"dirName ..."
	},

	{
		"more",	do_more,	2,	INFINITE_ARGS,
//		"fileName ..."
	},

	{
		"mv",		do_mv,		3,	INFINITE_ARGS,
//		"srcName ... destName"
	},

	{
		"pwd",		do_pwd,		1,	1,
//		""
	},

	{
		"rm",		do_rm,		2,	INFINITE_ARGS,
//		"fileName ..."
	},

	{
		"rmdir",	do_rmdir,	2,	INFINITE_ARGS,
//		"dirName ..."
	},
	{
		"source",	do_source,	2,	2,
//		"fileName"
	},

	#ifndef CBMSH_MINIMAL
	{
		"sum",		do_sum,		2,	INFINITE_ARGS,
//		"fileName ..."
	},
	#endif

	#ifndef CBMSH_MINIMAL
	{
		"unalias",	do_unalias,	2,	2,
//		"name"
	},
	#endif

	{
		"unit",		do_unit,		1,	2,
//		"[deviceNumber]"
	},

	{
		NULL,		0,		0,	0,
//		NULL
	}
};


// The definition of an command alias.
#ifndef CBMSH_MINIMAL
typedef struct {
	char *name;
	char *value;
} Alias;

static	Alias *aliasTable = NULL;
static	int	aliasCount;
#endif


// Local procedures.
static	void	readFile(const char * name);
static	void	command(const char * cmd);
static	BOOL	tryBuiltIn(const char * cmd);
static	void	runCmd(const char * cmd);
#ifndef CBMSH_MINIMAL
static	Alias *	findAlias(const char * name);
#endif

// Global interrupt flag.
//BOOL	intFlag;
char displayWidth, displayHeight;

/*
void perror (const char* s) {
	cprintf("%s: %s", s, strerror(errno));
}

char *cgets(char *str, char maxlen) {
	char c, len = 0, *buf = str;
	do {
		c = cgetc();
		cputc(c);
		*buf++ = c;
		len++;
	} while (c != '\n' && len < maxlen);
	*buf = '\0';
	cputc('\r');
	return str;
}
*/


void no_ram(const char *str) {
	printf("No ram for %s\n", str);
}


int main() {
	screensize(&displayWidth, &displayHeight);
	printf(GREETING);
	readFile(RC_FILENAME);
	// Read commands from stdin
	readFile(NULL);
	return 0;
}


// Read commands from the specified file. A null name pointer indicates stdin.
static void readFile(const char *name) {
	unsigned char cc;
	FILE *fp = stdin;
	static char buf[CMD_LEN];

	if (name)	{
		fp = fopen(name, "r");
		if (fp == NULL) {
			if (strcmp(name, RC_FILENAME)) perror(name);
			return;
		}
	}

	while (TRUE) {
		if (!name) putchar('>');

		if (fgets(buf, CMD_LEN, fp) == NULL) {
			if (ferror(fp) && errno == EINTR) {
				clearerr(fp);
				continue;
			}
			break;
		}
		if (!name) putchar('\n');

		if (cc = strlen(buf)) {
			if (buf[cc - 1] == '\n') cc--;
			while (cc > 0 && isBlank(buf[cc - 1])) cc--;
			buf[cc] = '\0';
			command(buf);
		}
	}

	if (ferror(fp)) {
		perror("cmdline");
		if (!name) exit(1);
	}

	clearerr(fp);
	if (name) fclose(fp);
}


/*
 * Parse and execute one null-terminated command line string.
 * This breaks the command line up into words, checks to see if the
 * command is an alias, and expands wildcards.
 */
static void command(const char *cmd) {
	#ifndef CBMSH_MINIMAL
	const Alias *alias;
	static char newCommand[CMD_LEN];
	#endif
	static char cmdName[CMD_LEN];
	const char *endCmd;

	/*
	 * Rest the interrupt flag and free any memory chunks that
	 * were allocated by the previous command.
	 */
	//intFlag = FALSE;
	//freeChunks();

	// Skip leading blanks.
	while (isBlank(*cmd)) cmd++;

	// If the command is empty or is a comment then ignore it.
	if (*cmd == '\0' || *cmd == '#') return;

	// Look for the end of the command name and then copy it to a buffer so we can null terminate it.
	endCmd = cmd;
	while (*endCmd && !isBlank(*endCmd)) endCmd++;
	memcpy(cmdName, cmd, endCmd - cmd);
	cmdName[endCmd - cmd] = '\0';

	/*
	 * Search for the command name in the alias table.
	 * If it is found, then replace the command name with
	 * the alias value, and append the current command
	 * line arguments to that.
	 */

	#ifndef CBMSH_MINIMAL
	alias = findAlias(cmdName);

	if (alias) {
		strcpy(newCommand, alias->value);
		strcat(newCommand, endCmd);
		cmd = newCommand;
	}
	#endif

	// Now look for the command in the builtin table, and execute if found.
	if (!tryBuiltIn(cmd)) runCmd(cmd);
	freeChunks();
}


/*
 * Try to execute a built-in command.
 * Returns TRUE if the command is a built in, whether or not the
 * command succeeds.  Returns FALSE if this is not a built-in command.
 */
static BOOL tryBuiltIn(const char *cmd) {
	const CommandEntry *entry;
	int argc;
	const char **argv, *endCmd = cmd;
	static char cmdName[CMD_LEN];

	// Look for the end of the command name and then copy the command name to a buffer so we can null terminate it.
	while (*endCmd && !isBlank(*endCmd)) endCmd++;
	memcpy(cmdName, cmd, endCmd - cmd);
	cmdName[endCmd - cmd] = '\0';

	// Search the command table looking for the command name.
	for (entry = commandEntryTable; entry->name != NULL; entry++) {
		if (!strcmp(entry->name, cmdName)) break;
	}

	// If the command is not a built-in, return indicating that.
	if (entry->name == NULL) return FALSE;

	// The command is a built-in. Break the command up into arguments and expand wildcards.
	if (!makeArgs(cmd, &argc, &argv)) return TRUE;

	// Give a usage string if the number of arguments is too large or too small.
	if ((argc < entry->minArgs) || (argc > entry->maxArgs)) {
//		printf("usage %s %s\n", entry->name, entry->usage);
		printf("Bad args\n");
		return TRUE;
	}

	// Call the built-in function with the argument list.
	entry->func(argc, argv);
	return TRUE;
}


// Execute the specified command
static void runCmd(const char * cmd) {
	int fp = open(cmd, O_RDONLY);
	BOOL isCart;
	unsigned int load_addr;

	if (fp <= 0) {
		printf("Can't run prg\n");
		return;
	}

	if (read(fp, &load_addr, 2) != 2) {
		printf("Can't run prg\n");
		close(fp);
		return;
	}
	close(fp);

	isCart = (load_addr > 8191);

	#if defined(__VIC20__)
		if (load_addr != 4609 && !isCart) {				// 4097 = unexpanded, 1025 = +3k
			POKE(648, 30);							// Screen memory page
			asm("jsr $E518");						// Initialize I/O
			POKEW(641, load_addr - 1);	// Start of memory for op system
			POKEW(643, 7680);						// Top of memory for op system
			POKEW(43, load_addr);				// Start of basic
			POKEW(55, 7680);						// Limit of memory for basic
			POKE(load_addr - 1, 0);			// won't run without this
		}
	#endif

	// 8E switches character set to 1 (default uppercase)
	printf("\x93\x8E\n\nload\"%s\",%u,1\x13", cmd, cbm_curunit());
	if (isCart) {
		memcpy((void*)KEYBUF_ADDR, "\nsys" RESET_SR "\n", 10);
		POKE(KEYBUF_NUMB, 10);
	} else {
		memcpy((void*)KEYBUF_ADDR, "\nrun\n", 5);
		POKE(KEYBUF_NUMB, 5);
	}
	exit(0);

	/*
	#if defined(__VIC20__)
		unsigned int p = 4613, size = 4607;
	#elif defined(__C64__)
		unsigned int p = 2053, size = 2047;
	#endif

	int fp = open(cmd, O_RDONLY), n;
	char buf[BUF_SIZE];

	if (fp <= 0) {
		printf("Can't run prg\n");
		return;
	}

	while ((n = read(fp, buf, sizeof(buf))) > 0) size += n;
	close(fp);

	getcwd(buf, BUF_SIZE);
	POKEW(p, 147 + 34 * 256);		// load + "
	p += 2;
	memcpy((void*)p, cmd, strlen(cmd));
	p += strlen(cmd);
	POKEW(p, 34 + 44 * 256);		// " + ,
	p += 2;
	memcpy((void*)p, buf, strlen(buf));
	p += strlen(buf);
	POKE(p++, 0); POKEW(p, 0);	// prg end
	//sprintf((char*)4613, "\x93\"%s\",%s\0\0\0", cmd, buf);  bigger!

	//asm("lda #00; jsr 65424");	// turn off control messages and kernal messages (looks like no effect)
	//asm("jsr 64909");		// initialise system constants
	//asm("jsr 64850");		// restore kernal vectors
	//asm("jsr _exit");		// cc65 cleanup

	//POKE(40320, 193);		// set 32k (megacart?)

	POKEW(45, size);	// set free ram (PEEK(34))

	#if defined(__VIC20__)
		asm("jsr $E518");		// Initialize I/O
		asm("jsr $E45B");		// initialize vectors
		asm("jsr $E3A4");		// initialize basic ram
		asm("jsr $C659");		// clr + reset stack (50765 new)
		asm("jsr $C533");		// relink basic code
		asm("jmp $C7AE");		// run (64802 reset)
	#elif defined(__C64__)
		POKEW(47, size);
		POKEW(49, size);
		asm("jsr $E518");		// Initialize I/O
		asm("jsr $E453");		// initialize vectors
		asm("jsr $E3BF");		// initialize basic ram
		//POKE(2050, 1);

		// currently doesn't work: looks like it crashes before doing the relink
		// Mike denial: If you set the CPU port register $01 to the standard value (#$37)
		// between these two groups of JSRs, it should work.

		asm("jsr $A659");		// clr + reset stack ($A65E or $A660, $A644 new)
		//asm("jsr $A68E");		// Set CHRGET pointer to BASIC start
		asm("jsr $A533");		// relink basic code
		asm("jmp $A7AE");		// run ($A871 is run command) lda #0;
	#endif
	*/
}


void do_help(int argc, const char **argv) {
	const CommandEntry *entry;
	// const char *str = NULL;

	// if (argc == 2) str = argv[1];
	/*
	if (str) {
		for (entry = commandEntryTable; entry->name; entry++) {
			if (!strcmp(str, entry->name)) {
				printf("usage %s %s\n", entry->name, entry->usage);
				return;
			}
		}
	}
	*/

	for (entry = commandEntryTable; entry->name; entry++) {
		// if ((argc == 1) || (strstr(entry->name, argv[1]) != NULL)) printf("%s\n", entry->name);
		printf(entry->name);
		putchar('\n');
	}
}


#ifndef CBMSH_MINIMAL
void do_alias(int argc, const char **argv) {
	static char buf[CMD_LEN];
	const char *name = argv[1];
	char *value;
	Alias *alias;
	int count;

	if (argc == 1) {
		count = aliasCount;
		for (alias = aliasTable; count--; alias++) printf("%s %s\n", alias->name, alias->value);
		return;
	}

	if (argc == 2) {
		alias = findAlias(name);
		if (alias) printf(alias->value);
		else printf("No such alias");
		putchar('\n');
		return;
	}

	if (!strcmp(name, "alias")) {
		printf("Invalid\n");
		return;
	}

	if (!makeString(argc - 2, argv + 2, buf, CMD_LEN)) return;

	value = malloc(strlen(buf) + 1);
	if (value == NULL) {
		no_ram("alias");
		return;
	}

	strcpy(value, buf);
	alias = findAlias(name);

	if (alias) {
		free(alias->value);
		alias->value = value;
		return;
	}

	if ((aliasCount % ALIAS_ALLOC) == 0) {
		count = aliasCount + ALIAS_ALLOC;

		if (aliasTable) alias = (Alias *) realloc(aliasTable, sizeof(Alias) * count);
		else alias = (Alias *) malloc(sizeof(Alias) * count);

		if (alias == NULL) {
			free(value);
			no_ram("alias tab");
			return;
		}

		aliasTable = alias;
	}

	alias = &aliasTable[aliasCount];
	alias->name = malloc(strlen(name) + 1);

	if (alias->name == NULL) {
		free(value);
		no_ram("alias name");
		return;
	}

	strcpy(alias->name, name);
	alias->value = value;
	aliasCount++;
}


// Look up an alias name, and return a pointer to it. Returns NULL if the name does not exist.
static Alias *findAlias(const char * name) {
	Alias *alias;
	int	count = aliasCount;
	for (alias = aliasTable; count--; alias++) {
		if (!strcmp(name, alias->name)) return alias;
	}
	return NULL;
}


void do_unalias(int argc, const char **argv) {
	Alias *alias;
	while (--argc) {
		alias = findAlias(*++argv);
		if (alias == NULL) continue;		// doesn't print any error message!

		free(alias->name);
		free(alias->value);
		aliasCount--;
		alias->name = aliasTable[aliasCount].name;
		alias->value = aliasTable[aliasCount].value;
	}
}
#endif


void do_source(int argc, const char **argv) {
	readFile(argv[1]);
}
